/******************************************************************
 **  Title             : LOD2DATA
 **
 **  System            : MS-DOS
 **
 **  Language          : Borland "C"
 **
 **  Author            : Bent Bagger
 **                      Bregnerdvej 151
 **                      DK-3460 Birkerd
 **                      Copyright (c) 1998
 **
 **  Date              : 1998-02-11
 **
 **  Version           : 1.0
 **
 **  Abstract          : Program to bring a .lod file onto assembly 
 **                      language form so that it can be edited with 
 **                      a text editor and reassembled.
 **                      The output is written on stdout
 **
 **
 ******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>


/* type definitions */

typedef unsigned char byte ;
typedef unsigned int int16;	/* 16-bit unsigned integer */


/* Function prototypes */

void do_end_processing (void) ;
void asm_data (byte *, int16) ;
void fmtline(int16, char *, int16) ;
void ctohex(char *, int16) ;
int  htoi(char *) ;
void putText (void) ;
void hex_dump (unsigned char *cp, int16 length) ;

byte *image ; /* binary image */
int16 im_start, im_offset ;
int16 modem_type, start_addr  ;
int16 debug = 0 ; /* The bits in 'debug' are used as follows:
		     0000 0000 0000 0001 : table info when reading .obj file
		   */

#define version_string ".lod file on .asm form. V1.0 - by OZ6BL\n"
char usage_string[] = { "Usage: lod2data [-v] [-h] -m modem type [-d <debug>] infile.ext\n \
   Output to stdout\n"} ;

void main (int argc, char **argv) {
	int n, type, line_length, length ;
	FILE *infp ;
	char debug_flags[100] = {'\0'}, modem_string[100] = {'\0'}, line_buf[100], *line_ptr ;
	int16 offset56, no_bytes ;
	int mem_type ;
	char mem_char ;
	byte *p ;
	int16 byte1, byte2, byte3, *pint16 ;

	/* Process command line options */
	if (argc < 3) {
		printf ("%s", usage_string) ;
		exit (0) ; }
	for (n = 1; n < argc; n++) {
		if (! strcmp (argv [n], "-d")) {strcpy (debug_flags,  argv [++n]); continue;}
		if (! strcmp (argv [n], "-v")) {printf (version_string) ; exit (0) ; }
		if (! strcmp (argv [n], "-m")) {strcpy (modem_string, argv [++n]); continue;}
		if (! strcmp (argv [n], "-h")) {printf ("%s", usage_string) ; exit (0) ;}
		if ((infp = fopen (argv[n], "rb")) == NULL) {
			perror (argv[n]) ;
			exit (1) ; }
	}

	/* Convert string values to numbers */
	if (debug_flags [0] != '\0') {
		debug = htoi(debug_flags); }
	if (modem_string[0] != '\0') {
		modem_type = htoi (modem_string) ; }
	/*
	   Allocate memory for 'image'
	*/
	if ((image = (byte *)malloc(10000)) == NULL) {
		fprintf (stderr, "Could not get enough memory\n") ;
		exit (1) ; }
	/*
	   Read the file into 'image'
	*/
	no_bytes = 0 ;
	im_start = 0 ;
	im_offset = 5 ; /* point after the offset field */
	start_addr = offset56 = 0 ;
	mem_type = -1 ;

	do {
		line_ptr = fgets (line_buf, 100, infp) ;
		line_length = strlen (line_buf) - 2 ; /* discard \r\n */
		line_buf[line_length] = '\0' ;
		if (line_length == 0) continue ; /* skip empty lines */
		if (line_buf[0] == '_') {
			/* special line */
			if (no_bytes != 0) {
				/* set memory type, offset, and length */
				if (mem_type >= 0 )
					image[im_start] = mem_type ;
				else {
					fprintf (stderr, "No _DATA line encountered\n") ;
					exit (1) ; }
				pint16 = (int16 *)&image[im_start+1] ;
				*pint16 = offset56 ;
				pint16  = (int16 *)&image[im_start+3] ;
				*pint16 = (no_bytes/3) ;
				if (debug & 1) hex_dump (&image[im_start], no_bytes + 5) ;
				/* length in 24 bit words */
				im_start += 5 + no_bytes ;
                                im_offset += 5 ;
				mem_type = -1 ;
				offset56 = no_bytes = 0 ; }
			switch (line_buf[1]) {
				case 'D' : {
					/* _DATA line - defines memory type and offset */
					sscanf (line_buf, "_DATA %c %x", &mem_char, &offset56) ;
					switch (mem_char) {
						case 'P' : mem_type = 0 ; break ;
						case 'X' : mem_type = 1 ; break ;
						case 'Y' : mem_type = 2 ; break ;
                                                default  : mem_type = -1 ; }
					break ; }
				case 'E' : {
					/* _END line - defines starting address for program */
					sscanf (line_buf, "_END %x", &start_addr) ;
					image[im_start] = 0xff ; /* insert end marker */
					do_end_processing () ;
					exit (0) ; /* end the program here */ }
				default: {
					fprintf (stderr, "Unexpected line encountered\n\t%s", line_buf) ;
					exit (1) ; } } }
		else {
			/* line with hex data */
			while (line_length != 0) {
				sscanf (line_ptr, "%02x%02x%02x ", &byte1, &byte2, &byte3) ;
				image [im_offset] = byte1 ;
				image [im_offset+1] = byte2 ;
				image [im_offset+2] = byte3 ;
				line_ptr += 7 ;
				line_length -= 7 ;
				no_bytes += 3 ;
				im_offset += 3 ; }
		}
	}
	while (line_ptr != NULL ) ;
}

void do_end_processing (void) {
	int16 offset = 0, length, offset56 ;
	byte *p ;
	char mem_type [10] ;

	printf ("\tHeader\t0%04xh,%04x,%04x,%04x,0%04xh\n",
		modem_type, 0, 0, 0, start_addr) ;
	printf ("\t\tPointer m%04x\n\t\tPointer 0\n", modem_type) ;
	printf ("m%04x:\n", modem_type) ;
	p = &image[offset] ;
	while (*p != 0xff) {
		offset56 = *(int16 *)(p+1) ;
		length = *(int16 *)(p+3) ;
		switch (*p) {
			case 0: { strcpy (mem_type, "P-data") ; break ; }
			case 1: { strcpy (mem_type, "X-data") ; break ; }
			case 2: { strcpy (mem_type, "Y-data") ; break ; }
			default: { strcpy (mem_type, "!!!-data") ; break ; } }
		printf (" db %02x\t; %s \n dw 0%xh\t; offset\n dw 0%xh\t; length\n",
			 *p, mem_type, offset56, length);
		asm_data(p+5, length) ;
		p += 5 + 3*length ;
	}
	printf (" db 0ffh\n") ;
}

void asm_data (byte *p, int16 length) {
/*
 * Print a buffer formatted as a *.lod file, i.e. a number
 * of 24 bit words. 'Length' is the number of words.
 */
   void asm_line (byte *, int16) ;
   int16 count = 0, n ;

   while (count < length) {
      n = (length-count) < 4 ? length-count : 4 ;
      asm_line (p, n) ;
      count += n ;
      p += 3*n ;
   }
}

void asm_line (byte *p, int16 n) {
/* Print a buffer up to 4 words (12 bytes) long in formatted hex
 *
 * db 030h,031h,032h, 033h,034h,035h, 036h,037h,038h, 039h,03ah,03bh
 */
   int16 c, i ;

   printf (" db") ;
   while (n-- != 0) {
      printf (" 0%02xh,0%02xh,0%02xh", *p & 0xff, *(p+1) & 0xff, *(p+2) & 0xff) ;
      p += 3 ;
      if (n != 0) printf (",") ;
   }
   printf ("\n") ;
}

/* Convert hex-ascii to integer */
int htoi(char *s)
{
	int i = 0;
	char c;

	while((c = *s++) != '\0'){
		if(c == 'x')
			continue;	/* allow 0x notation */
		if('0' <= c && c <= '9')
			i = (i * 16) + (c - '0');
		else if('a' <= c && c <= 'f')
			i = (i * 16) + (c - 'a' + 10);
		else if('A' <= c && c <= 'F')
			i = (i * 16) + (c - 'A' + 10);
		else
			break;
	}
	return i;
}

/* The following dump routines are adopted from KA9Q's
   NET. Give credit.....
*/


/* Extract a byte from a short */
#define	hibyte(x)	(((x) >> 8) & 0xff)
#define	lobyte(x)	((x) & 0xff)

/* Extract nibbles from a byte */
#define	hinibble(x)	(((x) >> 4) & 0xf)
#define	lonibble(x)	((x) & 0xf)

/* Dump an mbuf in hex */
void
hex_dump(unsigned char *cp, int16 length)
{
	int16 n;
	int16 address;
	void fmtline();

	address = 0;
	while(address < length) {
		n = (length-address) < 16 ? length-address : 16;
		fmtline(address,cp,n);
		address += n;
		cp += n ;
	}
}

/* Print a buffer up to 16 bytes long in formatted hex with ascii
 * translation, e.g.,
 * 0000: 30 31 32 33 34 35 36 37 38 39 3a 3b 3c 3d 3e 3f  0123456789:;<=>?
 */
void
fmtline(int16 addr, char *buf, int16 len)
{
	char line[78];
	register char *aptr,*cptr;
	register int16 c;

	memset(line,' ',sizeof(line));
	ctohex(line,(int16)hibyte(addr));
	ctohex(line+2,(int16)lobyte(addr));
	aptr = &line[6];
	cptr = &line[55];
	while(len-- != 0){
		c = *buf++ & 0xff;
		ctohex(aptr,c);
		aptr += 3;
		c &= 0x7f;
		*cptr++ = isprint(c) ? c : '.';
	}
	*cptr++ = '\0';
	printf("%s\n",line);
}
/* Convert byte to two ascii-hex characters */
static
void
ctohex(register char *buf, register int16 c)
{
	static char hex[] = "0123456789ABCDEF";

	buf[0] = hex[hinibble(c)];
	buf[1] = hex[lonibble(c)];
}